@@ -3,7 +3,7 @@
  * All Rights Reserved
  * **********************************************************/
 #ifdef VMX86_DEVEL
-char rcsId_vmx86[] = "$Id: vmx86.c,v 1.2 2003/02/16 01:29:50 bad Exp $";
+char rcsId_vmx86[] = "$Id: vmx86.c,v 1.4 2003/02/16 12:28:35 bad Exp $";
 #else
 #define FILECODE "F(301)"
 #endif 
@@ -110,6 +110,10 @@ Vmx86_Init(void *uniqHandle,    /* Unique id for VM to be created */
 #ifdef linux
    vm->logFD = -1;
 #endif
+#if defined(__NetBSD__) || defined(linux) && defined(KERNEL_2_4_0)
+   vm->startTime.time = HostIF_ReadTime(); 
+   vm->startTime.count = GET_TSC();
+#endif
    
    HostIF_GlobalVMLock(0);
    
