@@ -1,7 +1,7 @@
 /* **********************************************************
  * Copyright (C) 1998-2000 VMware, Inc.
  * All Rights Reserved
- * $Id: x86.h,v 1.3 2003/02/16 02:18:31 bad Exp $
+ * $Id: x86.h,v 1.4 2003/02/16 10:09:57 bad Exp $
  * **********************************************************/
 
 #ifndef _X86_H_
@@ -673,7 +673,10 @@ typedef struct ExcFrame {
 #define APIC_VTE_MASK         0x00010000
 #define APIC_VTE_TIMERMODE    0x00020000
 
-#define APIC_ID_MASK          0x0f000000
+/* Kernels 2.4.0-test8 and higher have the same definition */
+#ifndef APIC_ID_MASK
+#   define APIC_ID_MASK          0x0f000000
+#endif
 #define APIC_ID_SHIFT         24
 
 #define IOAPICID              0x00
@@ -833,5 +836,5 @@ typedef struct SMMStateSaveMap {
 
 #define SMM_RIGHTS_BITS   0x00f0ff00
 
-#endif _X86_H_
+#endif /* _X86_H_ */
 
