@@ -3,7 +3,7 @@
  * All Rights Reserved
  * **********************************************************/
 #ifdef VMX86_DEVEL
-char rcsId_driver[] = "$Id: driver.c,v 1.2 2003/02/16 01:29:51 bad Exp $";
+char rcsId_driver[] = "$Id: driver.c,v 1.3 2003/02/16 10:09:57 bad Exp $";
 #else
 #define FILECODE "F(300)"
 #endif 
@@ -31,6 +31,10 @@ extern int errno;
 #include <linux/interrupt.h>
 #endif
 
+#ifdef KERNEL_2_4_0
+#include <linux/spinlock.h>
+#endif
+
 #ifdef DO_APM
 #include <linux/apm_bios.h>
 static int LinuxDriverAPMCallback(apm_event_t event);
@@ -64,6 +68,8 @@ static int LinuxDriverAPMstate = APM_STATE_READY;
 #include "passthrough.h"	// _driver_ version of passthrough.h 
 #endif
 
+#include "vmmonInt.h"
+
 int errno;       // required for _exit()
 static void LinuxDriverQueue(VMLinux *vmLinux);
 static void LinuxDriverDequeue(VMLinux *vmLinux);
@@ -73,7 +79,7 @@ static Bool LinuxDriverCheckPadding(void);
 #ifdef VMX86_DEVEL
 #define TAGNAME "(development)"
 #else
-#define TAGNAME "$Name: vmware-netbsd-1-1 $"
+#define TAGNAME "$Name: vmware-netbsd-1-2 $"
 #endif
 
 /*
@@ -108,41 +114,13 @@ static int LinuxDriver_Close( struct inode *inode, struct file *filp );
 static unsigned int LinuxDriver_Poll(struct file *file, poll_table *wait);
 #else
 static void LinuxDriver_Close( struct inode *inode, struct file *filp );
-static int LinuxDriver_Select(struct inode *inode, struct file *filp,
-			      int sel_type, select_table *wait);
+static int LinuxDriver_Poll(struct inode *inode, struct file *filp,
+                            int sel_type, select_table *wait);
 #endif
 
 static void LinuxDriverSelectTimeout(VMLinux *vmLinux);
 
-#ifdef KERNEL_2_1
-static struct file_operations vmuser_fops = {
-   NULL,      /* lseek */
-   NULL,      /* read */
-   NULL,      /* write */
-   NULL,      /* readdir */
-   LinuxDriver_Poll,	/* select */
-   LinuxDriver_Ioctl,   /* ioctl */
-   NULL,      /* mmap */
-   LinuxDriver_Open,   /* just a selector for the real open */
-   NULL,		/* flush */
-   LinuxDriver_Close,   /* release */
-   NULL      /* fsync */
-};
-#else
-static struct file_operations vmuser_fops = {
-   NULL,      /* lseek */
-   NULL,      /* read */
-   NULL,      /* write */
-   NULL,      /* readdir */
-   LinuxDriver_Select,	/* select */
-   LinuxDriver_Ioctl,   /* ioctl */
-   NULL,      /* mmap */
-   LinuxDriver_Open,   /* just a selector for the real open */
-   LinuxDriver_Close,   /* release */
-   NULL      /* fsync */
-};
-#endif
-
+static struct file_operations vmuser_fops;
 
 
 /*
@@ -211,6 +189,16 @@ init_module(void)
    
    HostIF_InitSpinLock();
 
+   /* Initialize the file_operations structure. Because this code is always
+      compiled as a module, this is fine to do it here and not in a static
+      initializer --hpreg */
+   memset(&vmuser_fops, 0, sizeof(vmuser_fops));
+   compat_fop_set_owner(&vmuser_fops);
+   compat_fop_set_poll(&vmuser_fops, LinuxDriver_Poll);
+   vmuser_fops.ioctl = LinuxDriver_Ioctl;
+   vmuser_fops.open = LinuxDriver_Open;
+   vmuser_fops.release = LinuxDriver_Close;
+
 #ifdef VMX86_DEVEL
    devel_init_module();
    linuxState.minor = 0;
@@ -296,7 +284,9 @@ LinuxDriver_Open(struct inode * inode,
 {
    VMLinux *vmLinux;
    VMDriver *vm;
+#ifndef KERNEL_2_4_0
    uint32 flags;
+#endif
 
    if ((filp->f_flags & O_ACCMODE) == O_WRONLY) {
 #ifdef VMX86_DEVEL
@@ -306,26 +296,33 @@ LinuxDriver_Open(struct inode * inode,
 #endif
    }
 
+   compat_mod_inc_refcount;
    vmLinux = (VMLinux *) kmalloc(sizeof (VMLinux), GFP_KERNEL);
    if (vmLinux == NULL) {
       Warning("LinuxDriver_Open kmalloc failed\n");
+      compat_mod_dec_refcount;
       return -ENOMEM;
    }
    memset(vmLinux, 0, sizeof *vmLinux);
 #ifdef KERNEL_2_3_1
    init_waitqueue_head(&vmLinux->selectQueue);
 #endif
+#ifdef KERNEL_2_4_0
+   spin_lock_init(&vmLinux->selectLock);
+#endif
 
-   
+  
    vm = Vmx86_Init((void *)filp, (void *)(current->pid));
    if (vm == NULL) {
       Warning("Vmx86_Init failed\n");
       kfree(vmLinux);
+      compat_mod_dec_refcount;
       return -ENOMEM;
    }
    
    
       
+#ifndef KERNEL_2_4_0
    /*
     * Snap shot the time stamp counter and the real time so we
     * can later compute an estimate of the cycle time.
@@ -335,6 +332,7 @@ LinuxDriver_Open(struct inode * inode,
    vm->startTime.time = HostIF_ReadTime(); 
    vm->startTime.count = GET_TSC();
    RESTORE_FLAGS(flags);
+#endif
 
    vmLinux->vm = vm;
    vmLinux->selectTimer.data = (unsigned long) vmLinux;
@@ -352,7 +350,6 @@ LinuxDriver_Open(struct inode * inode,
 #endif
    
    //printk("/dev/vmmon: useCount INC pid=%d \n",current->pid);
-   MOD_INC_USE_COUNT;
 
 #ifdef DO_APM
    apm_register_callback(&LinuxDriverAPMCallback);
@@ -384,6 +381,19 @@ LinuxDriver_Close(struct inode * inode,
 {
    VMLinux *vmLinux = (VMLinux *) filp->private_data;
 
+#ifdef KERNEL_2_4_0
+retry:
+   spin_lock_bh(&vmLinux->selectLock);
+   if (vmLinux->selectWaiting) {
+      if (!del_timer(&vmLinux->selectTimer)) {
+         spin_unlock_bh(&vmLinux->selectLock);
+         goto retry;
+      }
+      vmLinux->selectWaiting = FALSE;
+   }
+   spin_unlock_bh(&vmLinux->selectLock);
+#endif
+
 #ifdef SUPPORT_LINUXVMWARE
    VMWare_SetVTracer(0);
 #endif
@@ -394,10 +404,6 @@ LinuxDriver_Close(struct inode * inode,
    } else {
       //printk("/dev/vmmon: useCount DEC pid=%d\n",vmLinux->pids[0]);
    }
-   
-   MOD_DEC_USE_COUNT;
-
-
 
    LinuxDriverDequeue(vmLinux);
    
@@ -414,6 +420,7 @@ LinuxDriver_Close(struct inode * inode,
 
    kfree(vmLinux);
    filp->private_data = NULL;
+   compat_mod_dec_refcount;
 #ifdef KERNEL_2_1
    return 0;
 #endif
@@ -423,7 +430,7 @@ LinuxDriver_Close(struct inode * inode,
 /*
  *----------------------------------------------------------------------
  *
- * LinuxDriver_Select  --
+ * LinuxDriver_Poll  --
  *
  *      This is only used to wake up the idle select() at the next
  *	clock tick.
@@ -431,7 +438,39 @@ LinuxDriver_Close(struct inode * inode,
  *
  *----------------------------------------------------------------------
  */
-#ifdef KERNEL_2_1
+#ifdef KERNEL_2_4_0
+unsigned int
+LinuxDriver_Poll(struct file *filp,
+		 poll_table *wait)
+{
+   VMLinux *vmLinux = (VMLinux *) filp->private_data;
+   int selectWaiting;
+   ASSERT(vmLinux);
+   if (wait != NULL) {
+retry:
+      spin_lock_bh(&vmLinux->selectLock);
+      if (vmLinux->selectWaiting) {
+         if(!del_timer(&vmLinux->selectTimer)) {
+            spin_unlock_bh(&vmLinux->selectLock);
+            goto retry;
+         }
+      } else {
+         vmLinux->selectWaiting = TRUE;
+      }
+      spin_unlock_bh(&vmLinux->selectLock);
+      poll_wait(filp, &vmLinux->selectQueue, wait);
+      init_timer(&vmLinux->selectTimer);
+      vmLinux->selectTimer.expires = jiffies + 1;
+      selectWaiting = vmLinux->selectWaiting;
+      add_timer(&vmLinux->selectTimer);
+   } else {
+      selectWaiting = vmLinux->selectWaiting;
+   }
+   return !selectWaiting;
+}
+
+#else 
+# ifdef KERNEL_2_1
 
 unsigned int 
 LinuxDriver_Poll(struct file *filp,
@@ -460,13 +499,13 @@ LinuxDriver_Poll(struct file *filp,
    return !vmLinux->selectWaiting;
 }
 
-#else		   
+# else		   
 
 int
-LinuxDriver_Select(struct inode *inode,
-                   struct file *filp,
-                   int sel_type,
-                   select_table *wait)
+LinuxDriver_Poll(struct inode *inode,
+                 struct file *filp,
+                 int sel_type,
+                 select_table *wait)
 {
 #if 0
    if (current->timeout) {
@@ -496,6 +535,7 @@ LinuxDriver_Select(struct inode *inode,
 #endif
 }
 
+# endif
 #endif
 /*
  *----------------------------------------------------------------------
@@ -510,9 +550,16 @@ LinuxDriver_Select(struct inode *inode,
 void
 LinuxDriverSelectTimeout(VMLinux *vmLinux)
 {
+#ifdef KERNEL_2_4_0
+   spin_lock(&vmLinux->selectLock);
+   vmLinux->selectWaiting = FALSE;
+   wake_up(&vmLinux->selectQueue);
+   spin_unlock(&vmLinux->selectLock);
+#else
    // We are already cli()
    vmLinux->selectWaiting = FALSE;
    wake_up(&vmLinux->selectQueue);
+#endif
 }
 
 
@@ -774,7 +821,10 @@ LinuxDriver_Ioctl( struct inode *inode,
       setVMPtr = TRUE;
 #endif
 #endif
-      ma = HostIF_APIC_Base(vm, setVMPtr);
+      /*
+       * hardcode probe to false in rel2_0
+       */
+      ma = HostIF_APIC_Base(vm, setVMPtr, /*probe*/ FALSE);                                             
       retval = ma;
       break;
    }
