@@ -38,6 +38,9 @@ typedef struct VMLinux {
    int numPendingPassthroughIRQs;
    unsigned char pendingPassthroughIRQs[NUM_PASSTHROUGH_IRQS];
 #endif
+#ifdef KERNEL_2_4_0
+   spinlock_t selectLock;
+#endif
 
 } VMLinux;
 
