@@ -3,7 +3,7 @@
  * All Rights Reserved
  * **********************************************************/
 #ifdef VMX86_DEVEL
-char rcsId_hostif[] = "$Id: hostif.c,v 1.2 2003/02/16 01:29:52 bad Exp $";
+char rcsId_hostif[] = "$Id: hostif.c,v 1.3 2003/02/16 10:09:58 bad Exp $";
 #else
 #define FILECODE "F(303)"
 #endif 
@@ -17,6 +17,7 @@ char rcsId_hostif[] = "$Id: hostif.c,v 1.2 2003/02/16 01:29:52 bad Exp $";
  */
 
 #include "driver-config.h"
+#include "hostif.h"  // must be 2nd (PAGE_SIZE)
 
 #include <linux/binfmts.h>
 #include <linux/fs.h>
@@ -48,10 +49,10 @@ char rcsId_hostif[] = "$Id: hostif.c,v 1.2 2003/02/16 01:29:52 bad Exp $";
 #include "vm_assert.h"
 #include "vm_asm.h"
 #include "modulecall.h"
-#include "hostif.h"
 #include "memtrack.h"
 #include "phystrack.h"
 
+#include "vmmonInt.h"
 
 
 /*
@@ -104,28 +105,19 @@ char rcsId_hostif[] = "$Id: hostif.c,v 1.2 2003/02/16 01:29:52 bad Exp $";
  */
 
 #ifdef CONFIG_HIGHMEM64G
-#error "64GB memory configuration is not supported."
+#   error "64GB memory configuration is not supported."
 #endif 
 
-#ifdef KERNEL_2_1
-#define ATOMIC_INCREMENT(_x) atomic_inc(&(_x))
-#define ATOMIC_DECREMENT(_x) atomic_dec(&(_x))
-#else
-#define ATOMIC_INCREMENT(_x) (_x)++
-#define ATOMIC_DECREMENT(_x) (_x)--
-#endif
 
+#define HOST_LOCK_PFN(_vm, _pfn) do {         \
+   get_page(mem_map + (_pfn));                \
+   PhysTrack_Add((_vm)->physTracker, (_pfn)); \
+} while (0)
 
-
-#define HOST_LOCK_PFN(_vm,_pfn)     {                   \
-   ATOMIC_INCREMENT((mem_map+(_pfn))->count);           \
-   PhysTrack_Add(_vm->physTracker,_pfn);                \
-}
-
-#define HOST_UNLOCK_PFN(_vm,_pfn)     {                 \
-   ATOMIC_DECREMENT((mem_map+(_pfn))->count);           \
-   PhysTrack_Remove(_vm->physTracker,_pfn);             \
-}
+#define HOST_UNLOCK_PFN(_vm, _pfn) do {          \
+   put_page(mem_map + (_pfn));                   \
+   PhysTrack_Remove((_vm)->physTracker, (_pfn)); \
+} while (0)
 
 #define HOST_ISTRACKED_PFN(_vm, _pfn) (PhysTrack_Test(_vm->physTracker, _pfn))
 
@@ -389,11 +381,19 @@ HostIF_LockPage(VMDriver *vm, void * addr)
       c = get_user((char *)addr); 
 #endif
 
-      /*
-       * now locate it and write the results back to the pmap
-       */
+        /*
+         * now locate it and write the results back to the pmap
+         *
+         * Under extreme memory pressure, the page may be gone again.
+         * Just fail the lock in that case.  (It was ASSERT_BUG(6339, mpn).)
+         * -- edward
+         */
+      
       mpn = FindMPN(vpn);
-      ASSERT(mpn);
+      if (mpn == 0) {
+         return 0;
+      }
+      
       
       /*
        * XXX SMP.
@@ -1069,6 +1069,213 @@ HostIF_GlobalVMUnLock(int callerID)
 }
 
 
+
+/*
+ * Misc. defined brought in from TOT (x86.h, vm_asm.h)
+ */
+
+#define CPUID_FAMILY_EXTENDED 15
+#define CPUID_EXTENDED_FAMILY_PENTIUM4	0
+#define CPUID_EXTENDED_FAMILY(_eax)   (((_eax) >> 20) & 0xff)
+#define CPUID_FAMILY_IS_486(_eax) (CPUID_FAMILY(_eax) == CPUID_FAMILY_486)
+#define CPUID_FAMILY_IS_P5(_eax)  (CPUID_FAMILY(_eax) == CPUID_FAMILY_P5)
+#define CPUID_FAMILY_IS_P6(_eax)  (CPUID_FAMILY(_eax) == CPUID_FAMILY_P6)
+#define CPUID_FAMILY_IS_PENTIUM4(_eax)  \
+           (CPUID_FAMILY(_eax) == CPUID_FAMILY_EXTENDED && \
+            CPUID_EXTENDED_FAMILY(_eax) == CPUID_EXTENDED_FAMILY_PENTIUM4)
+
+/*
+ * RDMSR/WRMSR access the 64bit MSRs as two
+ * 32 bit quantities, whereas GET_MSR/SET_MSR
+ * above access the MSRs as one 64bit quantity.
+ */
+
+#define RDMSR(msrNum, low, high) do { \
+   __asm__ __volatile__(              \
+      "rdmsr"                         \
+      : "=a" (low), "=d" (high)       \
+      : "c" (msrNum)                  \
+   );                                 \
+} while (0)
+ 
+#define WRMSR(msrNum, low, high) do { \
+   __asm__ __volatile__(              \
+      "wrmsr"                         \
+      : /* no outputs */              \
+      : "c" (msrNum),                 \
+        "a" (low),                    \
+        "d" (high)                    \
+    );                                \
+} while (0)
+     
+/*
+ * Utility routines for accessing and enabling the APIC
+ */
+
+#if defined(KERNEL_2_1) || defined(USE_PERFCOUNTERS)
+
+
+     
+/*
+ *----------------------------------------------------------------------
+ *
+ * GetMSR --
+ *
+ *      Wrapper for the macro that calls the rdmsr instruction.
+ *
+ * Results:
+ *      Returns the value of the MSR.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static INLINE uint64
+GetMSR(int index)
+{
+   uint64 msr;
+   msr = __GET_MSR (index);
+   return msr;
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * GetCPUID --
+ *
+ *      Wrapper for the macro that calls the cpuid instruction.
+ *
+ * Results:
+ *      Returns the value of cpuid in regs.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static INLINE void
+GetCPUID(int index, uint32 *regs)
+{
+   __GET_CPUID (index, regs);
+}
+
+/*
+ * Defines for accessing the APIC.  We use readl/writel to access the APIC
+ * which is how Linux wants you to access I/O memory (though on the x86
+ * just dereferencing a pointer works just fine).
+ */
+#define APICR_TO_ADDR(apic, reg)      (apic + (reg << 4))
+#define GET_APIC_REG(apic, reg)       (readl(APICR_TO_ADDR(apic, reg)))
+#define SET_APIC_REG(apic, reg, val)  (writel(val, APICR_TO_ADDR(apic, reg)))
+
+#define APIC_MAXLVT(apic)             ((GET_APIC_REG(apic, APICR_VERSION) >> 16) & 0xff)
+#define APIC_VERSIONREG(apic)         (GET_APIC_REG(apic, APICR_VERSION) & 0xff)
+
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * SetVMAPICPtr --
+ *
+ *      Sets a pointer to the APIC's virtual address in the VMDriver
+ *      structure.  
+ *
+ * Results:
+ *      None.
+ *
+ * Side effects:
+ *      The VMDriver structure is updated.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static void
+SetVMAPICPtr(VMDriver *vm, MPN mpn)
+{
+   volatile void *hostapic;
+
+   hostapic = (volatile void *) ioremap_nocache(MPN_2_MA(mpn), PAGE_SIZE);
+   if (hostapic) {
+      if ((APIC_VERSIONREG(hostapic) & 0xF0) == 0x10) {
+	 vm->hostAPIC = (volatile uint32 (*)[4]) hostapic;
+      }
+   }
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * ProbeAPIC --
+ *
+ *      Find the base physical address of the APIC.  On P6 family
+ *      processors, this is done by reading the address from an MSR.
+ *
+ * Results:
+ *      Returns the physical address of the APIC if found, 0 otherwise.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+ 
+MA
+ProbeAPIC(VMDriver *vm,       // IN/OUT: driver state
+	  Bool setVMPtr,      // IN: set a pointer to the APIC's virtual address
+	  Bool forceEnable)   // IN: APIC is being enabled here in the driver 
+{
+   uint32 regs[4];
+   uint32 *ptr;
+   uint32 features, version;
+   char name[16];
+   uint64 msr;
+   MPN mpn;
+
+   GetCPUID(0, regs);
+   ptr = (uint32 *)name;
+   ptr[0] = regs[1];
+   ptr[1] = regs[3];
+   ptr[2] = regs[2];
+   ptr[3] = 0;
+
+   if (strcmp(name,"GenuineIntel") == 0) {
+      GetCPUID(1, regs);
+      version = regs[0];
+      features = regs[3];
+
+      if ((features & CPUID_FEATURE_MSR) &&
+	  (forceEnable || (features & CPUID_FEATURE_APIC))) {
+
+	 /* APIC is present and enabled */
+         if (CPUID_FAMILY_IS_P6(version) ||
+             CPUID_FAMILY_IS_PENTIUM4(version)) {
+	    /* P6 allows APIC relocation */
+	    msr = GetMSR(APIC_BASE_MSR);
+	    mpn = (msr >> 12) & 0xFFFFFF;
+	    if (setVMPtr) {
+	       /* obtain a system address the APIC, only for P6.
+		  Not recommended for P5, per Intel Book 3, page 7-16 */
+	       SetVMAPICPtr(vm, mpn);
+	    } else {
+	       vm->hostAPIC = NULL;
+	    }
+	    return MPN_2_MA(mpn);
+	 } else if (!forceEnable) {
+	    return APIC_DEFAULT_ADDRESS;
+	 }
+      }
+   }
+   return 0;
+}
+#endif
+
+
 /*
  *----------------------------------------------------------------------
  *
@@ -1086,66 +1293,22 @@ HostIF_GlobalVMUnLock(int callerID)
  *
  *----------------------------------------------------------------------
  */
-/*
- * Utility routines 
- */
-#if defined(KERNEL_2_1) && defined(X86_APIC_PRESENT)
-#ifdef PROBE_APIC
-static uint64
-GetMSR(int index)
-{
-   uint64 msr;
-   msr = __GET_MSR (index);
-   return msr;
-}
-
-static void
-GetCPUID(int index, uint32 *regs)
-{
-   __GET_CPUID (index, regs);
-}
-#endif
-#endif
 
 MA
-HostIF_APIC_Base(VMDriver *vm, Bool setVMPtr)
+HostIF_APIC_Base(VMDriver *vm, Bool setVMPtr, Bool probe)
 {
-#if defined(KERNEL_2_1) && defined(X86_APIC_PRESENT)
+#if (defined(KERNEL_2_1) && defined(__SMP__)) || defined(CONFIG_X86_UP_IOAPIC) || defined(CONFIG_X86_UP_APIC)
 
    void *va;
    MPN mpn;
+   MA ma;
 
-#ifdef PROBE_APIC
-   uint32 regs[4];
-   uint32 *ptr;
-   char name[16];
-   uint8 family;
-   uint64 msr;
-
-   GetCPUID(0, regs);
-   ptr = (uint32 *)name;
-   ptr[0] = regs[1];
-   ptr[1] = regs[3];
-   ptr[2] = regs[2];
-   ptr[3] = 0;
-
-   if (strcmp(name,"GenuineIntel") == 0) {
-      GetCPUID(1, regs);
-      if ((regs[3] & CPUID_FEATURE_MSR) &&
-	  (regs[3] & CPUID_FEATURE_APIC)) {
-	 /* APIC is present and enabled */
-	 family = (regs[0] >> 8) & 0xF;
-	 if (family == 6) {
-	    /* P6 allows APIC relocation, encoded in MSR 1B */
-	    msr = GetMSR(0x1b);
-	    mpn = (msr >> 12) & 0xFFFFFF;
-	    return MPN_2_MA(mpn);
-	 } else {
-	    return APIC_DEFAULT_ADDRESS;
-	 }
+   if (probe) {
+      ma = ProbeAPIC(vm, setVMPtr, FALSE);
+      if (ma != 0) {
+	 return ma;
       }
    }
-#endif   
 
    va = (void *)__fix_to_virt(FIX_APIC_BASE);
    if (setVMPtr) {
@@ -1164,6 +1327,197 @@ HostIF_APIC_Base(VMDriver *vm, Bool setVMPtr)
 }
 
 
+#ifdef USE_PERFCOUNTERS
+
+/* 
+ * Defines for masking/unmasking PIC interrupts. 
+ */
+#define PIC_PRIMARY_MASK_PORT         0x21
+#define PIC_SECONDARY_MASK_PORT       0xA1
+
+#define MaskPICIRQs(mask1, mask2)        \
+   mask1 = INB(PIC_PRIMARY_MASK_PORT);   \
+   OUTB(PIC_PRIMARY_MASK_PORT, 0xff);    \
+   mask2 = INB(PIC_SECONDARY_MASK_PORT); \
+   OUTB(PIC_SECONDARY_MASK_PORT, 0xff);     
+
+#define UnmaskPICIRQs(mask1, mask2)      \
+   OUTB(PIC_PRIMARY_MASK_PORT, mask1);   \
+   OUTB(PIC_SECONDARY_MASK_PORT, mask2); 
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * HostIF_APICEnable --
+ *
+ *      Enable the APIC, if not already enabled.  The APIC is put into
+ *      through local mode.  In this mode the APIC passes interrupts 
+ *      along to the processor just as if they had come directly from the PIC.
+ *
+ * Results:
+ *      Returned value is APIC physical address.
+ *      
+ * Side effects:
+ *      APIC gets enabled.      
+ *
+ *----------------------------------------------------------------------
+ */
+
+MA 
+HostIF_APICEnable(VMDriver *vm)
+{
+   MA ma;
+   volatile void *va;
+   volatile uint8 picMask1, picMask2;
+   uint32 apicMSR, dummy;
+   uint32 tempMSR;
+   uint32 flags;
+
+   ASSERT(vm);
+   ASSERT(vm->hostAPIC == 0);
+
+   /*
+    *  Find the APIC's physical address. 
+    */
+   ma = ProbeAPIC(vm, FALSE, TRUE);
+   if (ma == 0) {
+      printk(KERN_WARNING "/dev/vmmon: Unable to find APIC's physaddr\n");
+      return 0;
+   }
+
+   /*
+    * Map APIC's physical address into a virtual address.
+    */
+   va = (volatile void *) ioremap_nocache(ma, PAGE_SIZE);
+   if (va == 0) {
+      printk(KERN_WARNING "/dev/vmmon: Unable to map APIC at physaddr 0x%x\n", ma);
+      return 0;
+   } 
+
+   /*
+    * Mask interrupts and enable the APIC.  Intel's Software Developer's Guide
+    * Guide Vol III, Appendix E says that we need to mask PIC interrupts before
+    * programming the local interrupt pins.
+    */
+   SAVE_FLAGS(flags); 
+   CLEAR_INTERRUPTS();
+   MaskPICIRQs(picMask1, picMask2);
+
+   RDMSR(APIC_BASE_MSR, apicMSR, dummy);
+   tempMSR = apicMSR;
+
+   if (!(apicMSR & APIC_MSR_ENABLED)) {
+      apicMSR |= APIC_MSR_ENABLED;
+
+      /*
+       * The enable bit of the APIC_BASE_MSR must be set before
+       * we can even read the APIC.
+       */
+      WRMSR(APIC_BASE_MSR, apicMSR, dummy);
+   }
+
+   /*
+    * Set up the local vector table entries.
+    */
+   SET_APIC_REG(va, APICR_SVR, APIC_SVR_ONES|APIC_SVR_VECTOR|APIC_SVR_APICENABLE);
+   SET_APIC_REG(va, APICR_LVT0, APIC_VTE_MODE_EXTINT);
+   SET_APIC_REG(va, APICR_LVT1, APIC_VTE_MODE_NMI);
+   SET_APIC_REG(va, APICR_ERRLVT, APIC_VTE_MASK);
+   SET_APIC_REG(va, APICR_TIMERLVT, APIC_VTE_MASK);
+
+   UnmaskPICIRQs(picMask1, picMask2);
+   RESTORE_FLAGS(flags); 
+
+   vm->hostAPIC =  (volatile uint32 (*)[4]) va;
+   return ma;
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * HostIF_APICSetPerfCtr --
+ *
+ *      Set the performance counter entry in the APIC. 
+ *
+ * Results:
+ *      TRUE if successful, FALSE otherwise.
+ *      
+ * Side effects:
+ *      APIC's PCLVT register is written.
+ *
+ *----------------------------------------------------------------------
+ */
+
+Bool
+HostIF_APICSetPerfCtr(VMDriver *vm,   // IN: driver state
+		      int gate)       // IN: interrupt gate for perfctr interrupts
+{
+   volatile void *apic;
+   uint32 maxlvt;
+
+
+   if (vm->hostAPIC == 0) {
+      printk(KERN_WARNING "PerfCtr: Could not set PCLVT, APIC not available\n");
+      return FALSE;
+   }
+
+   apic = (volatile void *) vm->hostAPIC;
+   maxlvt = APIC_MAXLVT(apic);
+
+   /*
+    * We need maximum LVT entries to be >= 4 in order for there to be
+    * an entry for the performance counters.
+    */
+   if (maxlvt >= 4) {
+      SET_APIC_REG(apic, APICR_PCLVT, APIC_VTE_MODE_FIXED | gate);
+   } else {
+      printk(KERN_WARNING "PerfCtr: Could not set PCLVT, APIC maxlvt %d < 4", maxlvt);
+      return FALSE;
+   }
+
+   return TRUE;
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * HostIF_APICAck --
+ *
+ *      Send an ack to the APIC
+ *
+ * Results:
+ *      None.
+ *      
+ * Side effects:
+ *      APIC's EOI register is written.
+ *
+ *----------------------------------------------------------------------
+ */
+
+void 
+HostIF_APICAck(VMDriver *vm)
+{
+   volatile void *apic;
+   volatile uint32 dummy;
+
+   apic = (volatile void *) vm->hostAPIC;
+
+   /*
+    * Send the APIC an EOI.  Dummy read is copied from Linux 
+    * ack_APIC_irq() routine. 
+    */
+   if (apic) {
+      dummy = GET_APIC_REG(apic, APICR_SVR);
+      SET_APIC_REG(apic, APICR_EOI, 0);
+   }
+}
+
+#endif // ifdef USE_PERFCOUNTERS
+
+
 /*
  *----------------------------------------------------------------------
  *
@@ -1189,7 +1543,8 @@ HostIF_APIC_Base(VMDriver *vm, Bool setVMPtr)
 MA
 HostIF_IOAPIC_Base(VMDriver *vm)
 {
-#if defined(KERNEL_2_1) && defined(X86_IOAPIC_PRESENT)
+#if (defined(KERNEL_2_1) && defined(__SMP__)) || defined(CONFIG_X86_UP_IOAPIC)
+
    int bus, slot, confword;
    uint32 id, apicbase;
    MPN mpn;
@@ -1223,4 +1578,3 @@ HostIF_IOAPIC_Base(VMDriver *vm)
 #endif
 }
 
-
