@@ -905,7 +905,7 @@ HostIF_GlobalVMUnLock(int callerID)
 }
 
 MA
-HostIF_APIC_Base(VMDriver *vm, Bool setVMPtr)
+HostIF_APIC_Base(VMDriver *vm, Bool setVMPtr, Bool probe)
 {
 	return 0;
 }
